unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  StdCtrls, ExtCtrls, Tabs, Preview, Dialogs, jpeg, ComCtrls;

type
  TMainForm = class(TForm)
    Toolbar: TPanel;
    ZoomComboBox: TComboBox;
    PrintButton: TButton;
    Label1: TLabel;
    Image1: TImage;
    PrintPreview: TPrintPreview;
    UnitComboBox: TComboBox;
    Label2: TLabel;
    FastPrintCheckBox: TCheckBox;
    Image2: TImage;
    AutoScrollCheckBox: TCheckBox;
    PageNavigator: TTabSet;
    PrinterSetupDialog: TPrinterSetupDialog;
    SaveButton: TButton;
    LoadButton: TButton;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    RichEdit1: TRichEdit;
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure ZoomComboBoxChange(Sender: TObject);
    procedure UnitComboBoxChange(Sender: TObject);
    procedure AutoScrollCheckBoxClick(Sender: TObject);
    procedure FastPrintCheckBoxClick(Sender: TObject);
    procedure PrintButtonClick(Sender: TObject);
    procedure SaveButtonClick(Sender: TObject);
    procedure LoadButtonClick(Sender: TObject);
    procedure PageNavigatorChange(Sender: TObject; NewTab: Integer;
      var AllowChange: Boolean);
    procedure PrintPreviewChange(Sender: TObject);
    procedure PrintPreviewBeforePrint(Sender: TObject);
    procedure PrintPreviewPrintProgress(Sender: TObject; PageNum,
      Progress: Integer; var AbortIt: Boolean);
    procedure PrintPreviewAfterPrint(Sender: TObject);
    procedure PrintPreviewNewPage(Sender: TObject);
    procedure PrintPreviewBeginDoc(Sender: TObject);
    procedure PrintPreviewEndDoc(Sender: TObject);
  private
    FirstActivation: Boolean;
    procedure CreateImageTextPage;
    procedure CreateImageOnlyPage;
    procedure CreateRichTextPage;
    procedure CreatePages;
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  Printers;

procedure TMainForm.FormCreate(Sender: TObject);
var
  SampleRTF: String;
begin
  Randomize;
  PrintPreview.ZoomState := zsZoomToFit;
  ZoomComboBox.ItemIndex := 6; // Zoom to Fit (Whole Page)
  UnitComboBox.ItemIndex := Ord(PrintPreview.Units);
  FastPrintCheckBox.Checked := PrintPreview.FastPrint;
  AutoScrollCheckBox.Checked := PrintPreview.AutoScroll;
  FirstActivation := True;
  SampleRTF := ExtractFilePath(Application.ExeName) + 'Sample.rtf';
  if FileExists(SampleRTF) then
    RichEdit1.Lines.LoadFromFile(SampleRTF);
end;

procedure TMainForm.FormActivate(Sender: TObject);
begin
  if FirstActivation then
  begin
    FirstActivation := False;
    Update;
    CreatePages;
  end;
end;

procedure TMainForm.ZoomComboBoxChange(Sender: TObject);
begin
  case ZoomComboBox.ItemIndex of
    0: PrintPreview.Zoom := 50;
    1: PrintPreview.Zoom := 100;
    2: PrintPreview.Zoom := 150;
    3: PrintPreview.Zoom := 200;
    4: PrintPreview.ZoomState := zsZoomToWidth;
    5: PrintPreview.ZoomState := zsZoomToHeight;
    6: PrintPreview.ZoomState := zsZoomToFit;
  end;
end;

procedure TMainForm.UnitComboBoxChange(Sender: TObject);
begin
  CreatePages;
end;

procedure TMainForm.AutoScrollCheckBoxClick(Sender: TObject);
begin
  PrintPreview.AutoScroll := AutoScrollCheckBox.Checked;
end;

procedure TMainForm.FastPrintCheckBoxClick(Sender: TObject);
begin
  PrintPreview.FastPrint := FastPrintCheckBox.Checked;
end;

procedure TMainForm.PrintButtonClick(Sender: TObject);
begin
  if (PrintPreview.State = psReady) and PrinterSetupDialog.Execute then
    PrintPreview.Print;
end;

procedure TMainForm.SaveButtonClick(Sender: TObject);
begin
  if SaveDialog.Execute then
  begin
    Screen.Cursor := crHourglass;
    Caption := Application.Title + ' - Saving to file...';
    try
      PrintPreview.SaveToFile(SaveDialog.FileName);
    finally
      Caption := Application.Title;
      Screen.Cursor := crDefault;
    end;
  end;
end;

procedure TMainForm.LoadButtonClick(Sender: TObject);
begin
  if OpenDialog.Execute then
  begin
    Screen.Cursor := crHourglass;
    Caption := Application.Title + ' - Loading from file...';
    try
      PrintPreview.LoadFromFile(OpenDialog.FileName);
    finally
      Caption := Application.Title;
      Screen.Cursor := crDefault;
    end;
  end;
end;

procedure TMainForm.PageNavigatorChange(Sender: TObject; NewTab: Integer;
  var AllowChange: Boolean);
begin
  PrintPreview.CurrentPage := NewTab + 1;
end;

procedure TMainForm.PrintPreviewChange(Sender: TObject);
begin
  while PageNavigator.Tabs.Count < PrintPreview.TotalPages do
    PageNavigator.Tabs.Add(IntToStr(PageNavigator.Tabs.Count + 1));
  while PageNavigator.Tabs.Count > PrintPreview.TotalPages do
    PageNavigator.Tabs.Delete(PageNavigator.Tabs.Count - 1);
  PageNavigator.TabIndex := PrintPreview.CurrentPage - 1;

  if PrintPreview.State = psReady then
  begin
    PrintButton.Enabled := PrintPreview.PrinterInstalled and (PrintPreview.TotalPages > 0);
    SaveButton.Enabled := (PrintPreview.TotalPages > 0);
  end;
end;

procedure TMainForm.PrintPreviewBeginDoc(Sender: TObject);
begin
  Caption := Application.Title + ' - Creating pages...';

  UnitComboBox.Enabled := False;
  PrintButton.Enabled := False;
  SaveButton.Enabled := False;
  LoadButton.Enabled := False;
end;

procedure TMainForm.PrintPreviewEndDoc(Sender: TObject);
begin
  Caption := Application.Title;

  UnitComboBox.Enabled := True;
  PrintButton.Enabled := PrintPreview.PrinterInstalled and (PrintPreview.TotalPages > 0);
  SaveButton.Enabled := (PrintPreview.TotalPages > 0);
  LoadButton.Enabled := True;
end;

procedure TMainForm.PrintPreviewBeforePrint(Sender: TObject);
begin
  Screen.Cursor := crHourglass;
  Caption := Application.Title + ' - Preparing to print...';

  UnitComboBox.Enabled := False;
  PrintButton.Enabled := False;
  SaveButton.Enabled := False;
  LoadButton.Enabled := False;
  FastPrintCheckBox.Enabled := False;
end;

procedure TMainForm.PrintPreviewAfterPrint(Sender: TObject);
begin
  Caption := Application.Title;
  Screen.Cursor := crDefault;

  UnitComboBox.Enabled := True;
  PrintButton.Enabled := PrintPreview.PrinterInstalled and (PrintPreview.TotalPages > 0);
  SaveButton.Enabled := (PrintPreview.TotalPages > 0);
  LoadButton.Enabled := True;
  FastPrintCheckBox.Enabled := True;
end;

procedure TMainForm.PrintPreviewPrintProgress(Sender: TObject; PageNum,
  Progress: Integer; var AbortIt: Boolean);
begin
  Caption := Format('%s - Printing page %d of %d (%%%d done)...',
    [Application.Title, PageNum, PrintPreview.TotalPages, Progress]);
  Update;
end;

// In this example, the code is independent of the Units property of
// PrintPreview. If you use only one measuremnt unit for PrintPreview, you can
// easily use constant values instead of passing them to ConvertUnit method.
// I also tried to write the code independent of the paper size.

procedure TMainForm.PrintPreviewNewPage(Sender: TObject);
var
  R: TRect;
  OneCM: Integer;
begin
  with PrintPreview do
  begin
    // The following line ensures one pixel pen width in any mapping mode.
    Canvas.Pen.Width := 0;
    Canvas.Brush.Style := bsCLear;
    // We are going to draw a rectangular frame on the page with 1cm distance
    // from edges of the paper. First, we convert 1cm to printer's current unit.
    OneCM := ConvertUnit(1000, mmHiMetric, Units);
    SetRect(R, 0, 0, PaperWidth, PaperHeight);
    InflateRect(R, -OneCM, -OneCM);
    Canvas.Rectangle(R.Left, R.Top, R.Right, R.Bottom);
    // Here we write the page number on bottom left corner of the page
    Canvas.Font.Height := -(OneCM div 2);   // Sets font's height to 50mm
    SetTextAlign(Canvas.Handle, TA_RIGHT or TA_TOP);
    Canvas.TextOut(R.Right, R.Bottom, Format('Page %d', [TotalPages+1]));
    SetTextAlign(Canvas.Handle, TA_LEFT or TA_TOP);
  end;
end;

procedure TMainForm.CreateImageTextPage;
var
  R: TRect;
  OneCM: Integer;
  SavedBottom: Integer;
begin
  with PrintPreview do
  begin
    // Don't forget that we have already drawn a frame for the page with 1cm
    // margin. We have to exclude it from the paper size. The content would
    // have 1cm margin relative to the frame.
    // First, we convert 1cm to printer's current unit.
    OneCM := ConvertUnit(1000, mmHiMetric, Units);
    SetRect(R, 0, 0, PaperWidth, PaperHeight);
    InflateRect(R, -(2 * OneCM), -(2 * OneCM));
    // We want to place an image horizontally in center of the paper and
    // under the frame's top. In addition, we want the image height does
    // not exceed 3cm limit.
    SavedBottom := R.Bottom;
    R.Bottom := R.Top + 3 * OneCM;
    PaintGraphicEx(R, Image1.Picture.Graphic, True, True, True);
    // We are going to draw a frame and write some text inside it. The new
    // frame is 1cm under the image boundary.
    R.Top := R.Bottom + OneCM;
    R.Bottom := SavedBottom;
    // Here, we draw the frame.
    Canvas.Rectangle(R.Left, R.Top, R.Right, R.Bottom);
    // To write the frame's dimensions under the frame, we use a font
    // with 0.5cm height.
    Canvas.Font.Height := -(OneCM div 2);
    Canvas.TextOut(R.Left, R.Bottom, Format('%d (%s) x %d (%s)',
      [R.Right - R.Left, UnitComboBox.Items[UnitComboBox.ItemIndex],
       R.Bottom - R.Top, UnitComboBox.Items[UnitComboBox.ItemIndex]]));
    // For the first line of the sample text, we set the font height to 0.8mm.
    Canvas.Font.Height := -ConvertUnit(80, mmHiMetric, Units);
    // While we have not reached to the frame's bottom...
    while R.Top - Canvas.Font.Height < R.Bottom do
    begin
      // Randomly we select a font color
      Canvas.Font.Color := RGB(Random(256), Random(256), Random(256));
      // draw the text,
      Canvas.TextRect(R, R.Left, R.Top, 'Powered by Borland Delphi.');
      // move the frame's top to the next line,
      Inc(R.Top, -Canvas.Font.Height);
      // and we increase the font height by 0.5mm.
      Canvas.Font.Height := Canvas.Font.Height - ConvertUnit(50, mmHiMetric, Units)
    end;
  end;
end;

procedure TMainForm.CreateImageOnlyPage;
var
  R: TRect;
  OneAndHalfCM: Integer;
begin
  with PrintPreview do
  begin
    // Don't forget that we have already drawn a frame for the page with 1cm
    // margin. We have to exclude it from the paper size. Besides that, we
    // want to include 0.5cm margin for the image also.
    // First, we convert 1.5cm to printer's current unit.
    OneAndHalfCM := ConvertUnit(1500, mmHiMetric, Units);
    SetRect(R, 0, 0, PaperWidth, PaperHeight);
    InflateRect(R, -OneAndHalfCM, -OneAndHalfCM);
    PaintGraphicEx(R, Image2.Picture.Graphic, True, False, True);
  end;
end;

procedure TMainForm.CreateRichTextPage;
var
  R: TRect;
  TwoCM: Integer;
begin
  with PrintPreview do
  begin
    // Don't forget that we have already drawn a frame for the page with 1cm
    // margin. We have to exclude it from the paper size. Besides that, we
    // want to include 1cm margin for the text also.
    // First, we convert 2cm to printer's current unit.
    TwoCM := ConvertUnit(2000, mmHiMetric, Units);
    SetRect(R, 0, 0, PaperWidth, PaperHeight);
    InflateRect(R, -TwoCM, -TwoCM);
    PaintRichText(R, RichEdit1, 0, nil);
  end;
end;

procedure TMainForm.CreatePages;
var
  I: Integer;
begin
  with PrintPreview do
  begin
    Units := TUnits(UnitComboBox.ItemIndex);
    BeginDoc;
    for I := 0 to 2 do
    begin
      if I > 0 then
        NewPage;
      case I mod 3 of
        0: CreateImageTextPage;
        1: CreateImageOnlyPage;
        2: CreateRichTextPage;
      end;
      Application.ProcessMessages;
    end;
    EndDoc;
  end;
end;

end.

