{

 This very simple component can be used to searching for files.
 File-masks (what to search) stored in the FilesToFind property.
 Initial directories (where to search) are stored in the StartDirs property.
 This unit is freeware. You can use, modify and distribute it freely.
 This unit was written to be helpfull to me and others.

 HANAX Software, 2003
 hanax@cenrum.sk

}


unit FindFile;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  EFindFileException = class(Exception);

  TFileWasFound = procedure(const fName, fPath: string) of object;


  TFindFile = class(TComponent)
  private
    FinF, FstDirs: TStringList;
    FIsSearching, Aborted: boolean;
    FTP: TThreadPriority;
    FOnStart: TNotifyEvent;
    FOnAbort: TNotifyEvent;
    FOnFound: TFileWasFound;
    FOnDone: TNotifyEvent;
    FDirStruct: TStringList;
    FOnSDDone: TNotifyEvent;
    procedure SetInF(const Value: TStringList);
    procedure SetStDirsF(const Value: TStringList);
    procedure Done;
    procedure FindIt;
    procedure Find;
    procedure BuildDirStruct(path: string);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Start;
    procedure Abort;
    property IsSearching: boolean read FIsSearching;
  published
    property FilesToFind: TStringList read FInF write SetInF;
    property StartDirs: TStringList read FstDirs write SetStDirsF;

    property OnStart: TNotifyEvent read FOnStart write FOnStart;
    property OnAbort: TNotifyEvent read FOnAbort write FOnAbort;
    property OnScanDirsDone: TNotifyEvent read FOnSDDone write FOnSDDone;
    property OnFound: TFileWasFound read FOnFound write FOnFound;
    property OnDone: TNotifyEvent read FOnDone write FOnDone;
  end;

procedure Register;

implementation

{ TFindFile }

procedure Register;
begin
  RegisterComponents('HANAX''s', [TFindFile]);
end;

procedure TFindFile.Start;
begin
  if FIsSearching then
    raise EFindFileException.Create('Searching in progress!');
  if FinF.Count = 0 then
    raise EFindFileException.Create('Nothing to search!');
  if FStDirs.Count = 0 then
    raise EFindFileException.Create('Nowhere to search!');
  FIsSearching := true;
  Aborted := false;
  if Assigned(FOnStart) then FOnStart(self);
  FindIt;
end;

procedure TFindFile.Abort;
begin
  FIsSearching := false;
  Aborted := true;
  if Assigned(FOnAbort) then FOnAbort(self);
end;

procedure TFindFile.Done;
begin
  FIsSearching := false;
  if Assigned(FOnDone) then FOnDone(self);
end;

constructor TFindFile.Create(AOwner: TComponent);
begin
  inherited;
  FinF := TStringList.Create;
  FstDirs := TStringList.Create;
  FDirStruct := TStringList.Create;
  FTP := tpNormal;
end;

destructor TFindFile.Destroy;
begin
  FinF.Free;
  FstDirs.Free;
  FDirStruct.Free;
  inherited;
end;

procedure TFindFile.SetInF(const Value: TStringList);
begin
  FinF.Assign(Value);
end;

procedure TFindFile.SetStDirsF(const Value: TStringList);
var
  a: integer;
  r: string;
begin
  FstDirs.Clear;
  for a := 0 to Value.Count - 1 do begin
    r := Value[a];
    if r[Length(r)] <> '\' then r := r + '\';
    FstDirs.Add(r);
  end;
end;


procedure TFindFile.BuildDirStruct(path: string);
var
  SR: TSearchRec;
begin
  if Aborted or not SetCurrentDir(path) then exit;
  FDirStruct.Add(path);
  Application.ProcessMessages;
  if FindFirst('*.*', faAnyFile, SR) = 0 then repeat
    if (SR.Name = '.') or (SR.Name = '..') then Continue;
    if SR.Attr and faDirectory > 0 then begin
      BuildDirStruct(path+SR.Name+'\');
      SetCurrentDir(path);
    end;
  until FindNext(SR) > 0;
  FindClose(SR);
end;


procedure TFindFile.Find;
var
  SR: TSearchRec;
  a: integer;
  b: integer;
begin
  for a := 0 to FDirStruct.Count - 1 do begin
    if aborted then break;
    SetCurrentDir(FDirStruct[a]);
    for b := 0 to FinF.Count - 1 do begin
      if aborted then break;
      if FindFirst(FinF[b], faAnyFile, SR) = 0 then repeat
        if aborted then break;
        Application.ProcessMessages;
        if (SR.Name = '.') or (SR.Name = '..') then Continue;
        if Assigned(FOnFound) then FOnFound(SR.Name, FDirStruct[a]);
      until FindNext(SR) > 0;
      FindClose(SR);
    end;
  end;
end;

procedure TFindFile.FindIt;
var
  a: integer;
begin
  for a := 0 to FstDirs.Count - 1 do BuildDirStruct(FstDirs[a]);
  if not aborted and Assigned(FOnSDDone) then FOnSDDone(self);
  if not aborted then Find;
  if not aborted then Done;
  FDirStruct.Clear;
end;

end.
