{
 -----------------------------------------------------------------------------

 MODUL          mdBUTTONS.pas

 COPYRIGHT      micro dynamics GmbH
 AUTHOR(S)      Michael Frank
 UPDATE         23.06.2003

 COMPILER       Delphi 5/6
 PLATFORM       Windows 95/98/Me/NT4/2000/XP

 DESCRIPTION    These components support Windows XP Visual Styles, if
                available.
                They are usable as direct replacements of the original
                Borland Delphi aequivalents.

 -----------------------------------------------------------------------------
 Revision History:   Original Version  04.03

 23.06.2003     Version 2.30
 -----------------------------------------------------------------------------
}
unit mdBUTTONS;

//============================================================================

interface

//============================================================================

uses
  Windows, Classes, Controls, CommCtrl, CheckLst, Forms, Graphics, Messages,
  ComCtrls, StdCtrls, SysUtils, ToolWin, Buttons,
  mdTHEMES;

type

{ TXPBitBtn }

  TXPBitBtn = class(TBitBtn)
  private
    fHasFocus : boolean;
    fMouseOver : boolean;
    procedure CMEnter (var Message : TCMGotFocus); message CM_ENTER;
    procedure CMExit (var Message : TCMExit); message CM_EXIT;
    procedure CMMouseEnter (var Message : TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave (var Message : TMessage); message CM_MOUSELEAVE;
    procedure CNDrawItem (var Message : TMessage); message CN_DRAWITEM;
    procedure WMEraseBkgnd (var Message : TWMEraseBkgnd); message WM_ERASEBKGND;
  public
    constructor Create (AOwner : TComponent); override;
  end;

{ TXPSpeedButton }

  TXPSpeedButton = class(TSpeedButton)
  private
    fBackgroundBuffer : TBitMap;
    fMouseOver : boolean;
    fParentColor : TColor;
    fParentLeft : integer;
    fParentHeight : integer;
    fParentTop : integer;
    fParentWidth : integer;
    fThemeChanged : boolean;
    fThemeName : WideString;
    fColorBuf : WideString;
    fSizeBuf : WideString;
    procedure CMEnabledChanged (var Message : TMessage); message CM_ENABLEDCHANGED;
    procedure CMMouseEnter (var Message : TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave (var Message : TMessage); message CM_MOUSELEAVE;
    procedure DestroyBackgroundBuffer;
  protected
    function IsThemeChanged : boolean; virtual;
    function IsParentChanged : boolean; virtual;
    procedure Paint; override;
  public
    constructor Create (AOwner : TComponent); override;
    destructor Destroy; override;
    procedure SetBounds (ALeft, ATop, AWidth, AHeight : integer); override;
  end;

//============================================================================

implementation

//============================================================================

//----------------------------------------------------------------------------
procedure CalcButtonLayout (Control : TControl; DC : HDC; BiDiMode : TBiDiMode;
  const ButtonRect : TRect; var GlyphPos : TPoint; var CaptionRect : TRect);
//----------------------------------------------------------------------------

var
  iMargin : integer;
  iSpacing : integer;
  iNumGlyphs : integer;
  iButtonWidth, iButtonHeight : integer;
  iCaptionWidth, iCaptionHeight : integer;
  iGlyphWidth, iGlyphHeight : integer;
  iWidth, iHeight : integer;
  Layout : TButtonLayout;
  CaptionPos : TPoint;
  Glyph : TBitmap;

begin
  if (Control is TBitBtn) then
  begin
    Glyph:= TBitBtn(Control).Glyph;
    Layout:= TBitBtn(Control).Layout;
    iMargin:= TBitBtn(Control).Margin;
    iSpacing:= TBitBtn(Control).Spacing;
    iNumGlyphs:= TBitBtn(Control).NumGlyphs;
  end else
  begin
    Glyph:= TSpeedButton(Control).Glyph;
    Layout:= TSpeedButton(Control).Layout;
    iMargin:= TSpeedButton(Control).Margin;
    iSpacing:= TSpeedButton(Control).Spacing;
    iNumGlyphs:= TSpeedButton(Control).NumGlyphs;
  end;

  if (BiDiMode = bdRightToLeft) then
  begin
    if (Layout = blGlyphLeft) then
      Layout:= blGlyphRight
    else if (Layout = blGlyphRight) then
      Layout:= blGlyphLeft;
  end;

  iButtonWidth:= ButtonRect.Right - ButtonRect.Left;
  iButtonHeight:= ButtonRect.Bottom - ButtonRect.Top;
  iCaptionWidth:= CaptionRect.Right - CaptionRect.Left;
  iCaptionHeight:= CaptionRect.Bottom - CaptionRect.Top;
  if Assigned(Glyph) then
  begin
    iGlyphWidth:= Glyph.Width div iNumGlyphs;
    iGlyphHeight:= Glyph.Height;
  end else
  begin
    iGlyphWidth:= 0;
    iGlyphHeight:= 0;
  end;

  if (Layout in [blGlyphLeft, blGlyphRight]) then
  begin
    GlyphPos.Y:= (iButtonHeight - iGlyphHeight) div 2;
    CaptionPos.Y:= (iButtonHeight - iCaptionHeight) div 2;
  end else
  begin
    GlyphPos.X:= (iButtonWidth - iGlyphWidth) div 2;
    CaptionPos.X:= (iButtonWidth - iCaptionWidth) div 2;
  end;

  if (iCaptionWidth = 0) or (iGlyphWidth = 0) then
    iSpacing:= 0;

  if (iMargin < 0) then
  begin
    if (iSpacing < 0) then
    begin
      iWidth:= iGlyphWidth + iCaptionWidth;
      iHeight:= iGlyphHeight + iCaptionHeight;

      if (Layout in [blGlyphLeft, blGlyphRight]) then
        iMargin:= (iButtonWidth - iWidth) div 3
      else
        iMargin:= (iButtonHeight - iHeight) div 3;

      iSpacing:= iMargin;
    end else
    begin
      iWidth:= iGlyphWidth + iCaptionWidth + iSpacing;
      iHeight:= iGlyphHeight + iCaptionHeight + iSpacing;

      if (Layout in [blGlyphLeft, blGlyphRight]) then
        iMargin:= (iButtonWidth - iWidth) div 2
      else
        iMargin:= (iButtonHeight - iHeight) div 2;
    end;
  end else

  begin
    if (iSpacing < 0) then
    begin
      iWidth:= iButtonWidth - iGlyphWidth - iMargin;
      iHeight:= iButtonHeight + iGlyphHeight - iMargin;

      if (Layout in [blGlyphLeft, blGlyphRight]) then
        iSpacing:= (iWidth - iCaptionWidth) div 2
      else
        iSpacing:= (iHeight - iCaptionHeight) div 2;
    end;
  end;

  case Layout of
  blGlyphLeft :
    begin
      GlyphPos.X:= iMargin;
      CaptionPos.X:= GlyphPos.X + iGlyphWidth + iSpacing;
    end;
  blGlyphRight :
    begin
      GlyphPos.X:= iButtonWidth - iMargin - iGlyphWidth;
      CaptionPos.X:= GlyphPos.X - iSpacing - iCaptionWidth;
    end;
  blGlyphTop :
    begin
      GlyphPos.Y:= iMargin;
      CaptionPos.Y:= GlyphPos.Y + iGlyphHeight + iSpacing;
    end;
  blGlyphBottom :
    begin
      GlyphPos.Y:= iButtonHeight - iMargin - iGlyphHeight;
      CaptionPos.Y:= GlyphPos.Y - iSpacing - iCaptionHeight;
    end;
  end;

  with GlyphPos do
  begin
    inc(X, ButtonRect.Left);
    inc(Y, ButtonRect.Top);
  end;

  OffsetRect(CaptionRect, CaptionPos.X, CaptionPos.Y);
end;

//----------------------------------------------------------------------------
function GetDisabledGlyph (Source : TBitmap) : TBitmap;
//----------------------------------------------------------------------------

var
  X, Y : integer;
  Color : TColor;
  R, G, B : integer;

begin
  Result:= TBitmap.Create;
  with Result do
  begin
    Assign(Source);
    X:= 0;
    while (X < Width) do
    begin
      Y:= 0;
      while (Y < Height) do
      begin
        Color:= Canvas.Pixels[X,Y];
        R:= Color and $0000FF;
        G:= Color and $00FF00 shr 8;
        B:= Color and $FF0000 shr 16;
{
        Color:= round(0.212671 * R + 0.715160 * G + 0.072169 * B);
}
        Color:= (R + G + B) div 3;
        Canvas.Pixels[X,Y]:= Color shl 16 + Color shl 8 + Color;
        inc(Y,1);
      end;
      inc(X,1);
    end;
  end;
end;

//----------------------------------------------------------------------------
procedure DrawButtonGlyph (Source : TBitmap; DC : HDC; SourcePos : TPoint;
  NumGlyphs, State : integer; Enabled : boolean);
//----------------------------------------------------------------------------

var
  i : integer;
  Destination : TBitmap;
  Mask : TBitmap;
  DestinationPos : TPoint;
  iGlyphWidth : integer;

begin
  Destination:= Source;
  DestinationPos.X:= 0;
  DestinationPos.Y:= 0;

  if not Source.Empty then
  begin
    iGlyphWidth:= Source.Width div NumGlyphs;

    if (NumGlyphs > 1) or Enabled then
    begin
      case State of
      PBS_NORMAL :
        i:= 1;
      PBS_PRESSED :
        i:= 3;
      PBS_DISABLED :
        i:= 4;
      PBS_HOT :
        i:= 2;
      else
        i:= 1;
      end;
      if (i > NumGlyphs) then
        i:= 1;
      DestinationPos.X:= pred(i) * iGlyphWidth;
    end else
      Destination:= GetDisabledGlyph(Source);

    if not Destination.Empty then
    begin
      Mask:= TBitmap.Create;
      try
        Mask.Assign(Source);
        Mask.Mask(Source.TransparentColor);

        TransparentStretchBlt(DC,
          SourcePos.X, SourcePos.Y, iGlyphWidth, Source.Height,
                              Destination.Canvas.Handle,
          DestinationPos.X, DestinationPos.Y, iGlyphWidth, Destination.Height,
                              Mask.Canvas.Handle,
          DestinationPos.X, DestinationPos.Y);
      finally
        Mask.Free;
      end;
    end;
  end;
end;

//============================================================================
// class TXPBitBtn
//============================================================================

//----------------------------------------------------------------------------
constructor TXPBitBtn.Create (AOwner : TComponent);
//----------------------------------------------------------------------------

begin
  inherited;
  fMouseOver:= false;
  fHasFocus:= false;

  DoubleBuffered:= true;
end;

//----------------------------------------------------------------------------
procedure TXPBitBtn.WMEraseBkgnd (var Message : TWMEraseBkgnd);
//----------------------------------------------------------------------------

begin
  if UseThemes and not (csDesigning in ComponentState) then
  begin
    DrawThemeParentBackground(Handle, Message.DC, nil);

    Message.Result:= 1;
  end else
    inherited;
end;

//----------------------------------------------------------------------------
procedure TXPBitBtn.CMEnter (var Message : TCMGotFocus);
//----------------------------------------------------------------------------

begin
  inherited;
  fHasFocus:= true;
end;

//----------------------------------------------------------------------------
procedure TXPBitBtn.CMExit (var Message : TCMExit);
//----------------------------------------------------------------------------

begin
  inherited;
  fHasFocus:= false;
end;

//----------------------------------------------------------------------------
procedure TXPBitBtn.CMMouseEnter (var Message : TMessage);
//----------------------------------------------------------------------------

begin
  inherited;

  if UseThemes and not (csDesigning in ComponentState) then
  begin
    fMouseOver:= true;
    Repaint;
  end;
end;

//----------------------------------------------------------------------------
procedure TXPBitBtn.CMMouseLeave (var Message : TMessage);
//----------------------------------------------------------------------------

begin
  inherited;

  if UseThemes and not (csDesigning in ComponentState) then
  begin
    fMouseOver:= false;
    Repaint;
  end;
end;

//----------------------------------------------------------------------------
procedure TXPBitBtn.CNDrawItem (var Message : TMessage);
//----------------------------------------------------------------------------

var
  hhTheme : HTHEME;
  iColor : TColor;
  iDrawState : integer;
  DrawRect : TRect;
  CaptionRect : TRect;
  GlyphPos : TPoint;
  s : WideString;

begin
  hhTheme:= 0;

  if UseThemes and not (csDesigning in ComponentState) then
    hhTheme:= OpenThemeData(0, 'Button');

  if (hhTheme <> 0) then
  with TWMDrawItem(Message).DrawItemStruct^ do
  try
    if Enabled then
    begin
      if Bool(ItemState and ODS_SELECTED) then
        iDrawState:= PBS_PRESSED
      else if fMouseOver then
        iDrawState:= PBS_HOT
      else if Self.Default or Focused then
        iDrawState:= PBS_DEFAULTED
      else
        iDrawState:= PBS_NORMAL;
    end else
      iDrawState:= PBS_DISABLED;

    if IsThemeBackgroundPartiallyTransparent(hhTheme, BP_PUSHBUTTON, iDrawState) then
      DrawThemeParentBackground(hwndItem, hDC, @rcItem);

    DrawThemeBackground(hhTheme, hDC, BP_PUSHBUTTON, iDrawState, @rcItem, nil);

    if fHasFocus then
    begin
      GetThemeColor(hhTheme, BP_PUSHBUTTON, iDrawState, TMT_TEXTCOLOR, iColor);
      SetTextColor(Handle, iColor);
      GetThemeBackgroundContentRect(hhTheme, hDC, BP_PUSHBUTTON, iDrawState,
        @rcItem, DrawRect);
      DrawFocusRect(hDC, DrawRect);
    end;

    s:= WideString(Caption);
    GetThemeTextExtent(hhTheme, hDC, BP_PUSHBUTTON, iDrawState,
      PWideChar(s), -1, 0, nil, CaptionRect);
    CalcButtonLayout(Self, hDC, BiDiMode, rcItem, GlyphPos, CaptionRect);

    DrawThemeText(hhTheme, hDC, BP_PUSHBUTTON, iDrawState,
      PWideChar(s), -1, 0, 0, @CaptionRect);

    DrawButtonGlyph(Glyph, hDC, GlyphPos, NumGlyphs, iDrawState, Enabled);
  finally
    CloseThemeData(hhTheme);
  end else
    inherited;
end;

//============================================================================
// class TXSpeedButton
//============================================================================

//----------------------------------------------------------------------------
constructor TXPSpeedButton.Create (AOwner : TComponent);
//----------------------------------------------------------------------------

begin
  inherited;
  Width:= 50;
  Height:= 25;
  fMouseOver:= false;

  Transparent:= false;
  fBackgroundBuffer:= nil;

  if UseThemes then
    IsThemeChanged;

  fThemeChanged:= false;
end;

//----------------------------------------------------------------------------
destructor TXPSpeedButton.Destroy;
//----------------------------------------------------------------------------

begin
  DestroyBackgroundBuffer;
  inherited;
end;

//----------------------------------------------------------------------------
procedure TXPSpeedButton.DestroyBackgroundBuffer;
//----------------------------------------------------------------------------

begin
  if (fBackgroundBuffer <> nil) then
  begin
    fBackgroundBuffer.Free;
    fBackgroundBuffer:= nil;
  end;
end;

//----------------------------------------------------------------------------
procedure TXPSpeedButton.CMEnabledChanged (var Message : TMessage);
//----------------------------------------------------------------------------

begin
  inherited;

  if not Transparent and not Flat and
     UseThemes and not (csDesigning in ComponentState) then
  begin
    if (Parent is TForm) then
    begin
      DestroyBackgroundBuffer;
      fThemeChanged:= true;
      Transparent:= true;
      Invalidate;
    end;
  end;
end;

//----------------------------------------------------------------------------
procedure TXPSpeedButton.CMMouseEnter (var Message : TMessage);
//----------------------------------------------------------------------------

begin
  inherited;

  if UseThemes and not (csDesigning in ComponentState) then
  begin
    fMouseOver:= true;
    Repaint;
  end;
end;

//----------------------------------------------------------------------------
procedure TXPSpeedButton.CMMouseLeave (var Message : TMessage);
//----------------------------------------------------------------------------

begin
  inherited;

  if UseThemes and not (csDesigning in ComponentState) then
  begin
    fMouseOver:= false;
    Repaint;
  end;
end;

//----------------------------------------------------------------------------
function TXPSpeedButton.IsThemeChanged : boolean;
//----------------------------------------------------------------------------

var
  achThemeName, achColor, achSize : array[0..1023] of WideChar;
  wsThemeName, wsColorBuf, wsSizeBuf : WideString;
  hr : HResult;

begin
  result:= false;

  if UseThemes and not (csDesigning in ComponentState) then
  begin
    hr:= GetCurrentThemeName(@achThemeName, 1024, @achColor, 1024, @achSize, 1024);
    if (hr = S_OK) then
    begin
      wsThemeName:= achThemeName;
      wsColorBuf:= achColor;
      wsSizeBuf:= achSize;
      if (wsThemeName <> fThemeName) or
         (wsColorBuf <> fColorBuf) or
         (wsSizeBuf <> fSizeBuf) then
      begin
        fThemeName:= wsThemeName;
        fColorBuf:= wsColorBuf;
        fSizeBuf:= wsSizeBuf;

        result:= true;
      end;
    end;
  end;
end;

//----------------------------------------------------------------------------
function TXPSpeedButton.IsParentChanged : boolean;
//----------------------------------------------------------------------------

begin
  result:= false;

  if UseThemes and not (csDesigning in ComponentState) then
  begin
    if (Parent is TForm) then
    with Parent as TForm do
    begin
      if (Color <> fParentColor) or
         (Width <> fParentWidth) or
         (Height <> fParentHeight) or
         (Self.Left <> fParentLeft) or
         (Self.Top <> fParentTop) then
      begin
        fParentColor:= Color;
        fParentWidth:= Width;
        fParentHeight:= Height;
        fParentLeft:= Self.Left;
        fParentTop:= Self.Top;

        result:= true;
      end;
    end;
  end;
end;

//----------------------------------------------------------------------------
procedure TXPSpeedButton.Paint;
//----------------------------------------------------------------------------

var
  hhTheme : HTHEME;
  iDrawState : integer;
  DrawRect : TRect;
  CaptionRect : TRect;
  GlyphPos : TPoint;
  Bmp : TBitmap;
  s : WideString;

begin
  hhTheme:= 0;

  if UseThemes and not (csDesigning in ComponentState) then
  begin
    if Flat then
      hhTheme:= OpenThemeData(0, 'Toolbar')
    else
      hhTheme:= OpenThemeData(0, 'Button');
  end;

  if (hhTheme <> 0) then
  try
    if not Transparent then
    begin
      if IsThemeChanged or IsParentChanged then
      begin
        DestroyBackgroundBuffer;
        fThemeChanged:= true;
        Transparent:= true;
        Repaint;
      end;
    end else
    if fThemeChanged then
    begin
      Transparent:= false;
      fThemeChanged:= false;
    end;

    if Enabled then
    begin
      if (Self.fState in [bsDown, bsExclusive]) then
        iDrawState:= PBS_PRESSED
      else if fMouseOver then
        iDrawState:= PBS_HOT
      else
        iDrawState:= PBS_NORMAL;
    end else
      iDrawState:= PBS_DISABLED;

    DrawRect:= Rect(0, 0, Width, Height);

    if (fBackgroundBuffer = nil) then
    begin
      fBackgroundBuffer:= TBitmap.Create;
      with fBackgroundBuffer do
      begin
        Width:= Self.Width;
        Height:= Self.Height;
        Canvas.CopyMode:= cmSrcCOPY;
        Canvas.CopyRect(DrawRect,Self.Canvas,DrawRect);
      end;
    end;

    Bmp:= TBitmap.Create;
    try
      with Bmp do
      begin
        Width:= Self.Width;
        Height:= Self.Height;
        Canvas.CopyMode:= cmSrcCOPY;
        if (fBackgroundBuffer <> nil) then
          Canvas.CopyRect(DrawRect,fBackgroundBuffer.Canvas,DrawRect);

        DrawThemeBackground(hhTheme, Canvas.Handle, BP_PUSHBUTTON, iDrawState, @DrawRect, nil);

        s:= WideString(Caption);
        SelectObject(Canvas.Handle, Font.Handle);
        GetThemeTextExtent(hhTheme, Canvas.Handle, BP_PUSHBUTTON, iDrawState,
          PWideChar(s), -1, 0, nil, CaptionRect);

        CalcButtonLayout(Self, Canvas.Handle, BiDiMode, DrawRect, GlyphPos, CaptionRect);

        DrawThemeText(hhTheme, Canvas.Handle, BP_PUSHBUTTON, iDrawState,
          PWideChar(s), -1, 0, 0, @CaptionRect);

        DrawButtonGlyph(Glyph, Canvas.Handle, GlyphPos, NumGlyphs, iDrawState, Enabled);
      end;
      Canvas.Draw(0, 0, Bmp);
    finally
      Bmp.Free;
    end;
  finally
    CloseThemeData(hhTheme);
  end else
    inherited;
end;

//----------------------------------------------------------------------------
procedure TXPSpeedButton.SetBounds (ALeft, ATop, AWidth, AHeight : integer);
//----------------------------------------------------------------------------

begin
  if ((ALeft <> Left) or (ATop <> Top) or
     (AWidth <> Width) or (AHeight <> Height)) then
  begin
    DestroyBackgroundBuffer;
  end;
  inherited;
end;

//============================================================================

end.



